import matplotlib.pyplot as plt
import mne
import pandas as pd
from mne.io import Raw


def good_epoch_dur(raw, seg_start, seg_end, seg_len):
    """
    Calculate the valid duration of an epoch, excluding bad segments.

    Returns
    -------
    float
        Valid duration of the epoch.
    """

    bad_dur = sum(max(0, min(anno['onset'] + anno['duration'], seg_end) - max(anno['onset'], seg_start))
                  for anno in raw.annotations if 'Bad_epoch' in anno['description'])

    return max(0, seg_len - bad_dur)


def set_up_raw(raw: Raw, rerefer: bool = False, chan_limit=None,
               montage_temp: str = "standard_1005", is_montage: bool = False, drop_chan=()) -> Raw:
    ch_drop = [
        ch for ch in raw.ch_names
        if ch.startswith('M') or 'EMG' in ch or 'EOG' in ch or 'ECG' in ch or 'chin' in ch.lower() or ch.startswith('E')
    ]
    raw.drop_channels(ch_drop)
    raw.drop_channels(drop_chan, on_missing='warn')
    if chan_limit is not None:
        raw = raw.pick_channels(chan_limit, ordered=False, on_missing='warn')
    if rerefer:
        raw.set_eeg_reference(ref_channels=['M1', 'M2'])
    if is_montage:
        montage = mne.channels.make_standard_montage(montage_temp)
        raw.set_montage(montage, on_missing='warn')
    return raw


def event_lock(raw, SO_candidates, SP_candidates, event_summary, window=1.5, verbose=True):
    """
    Identify spindles occurring within ±1.5 s of each detected SO trough.

    Parameters:
    - raw (mne.io.Raw): EEG raw data.
    - SO_candidates (pd.DataFrame): DataFrame containing detected SO events.
    - SP_candidates (pd.DataFrame): DataFrame containing detected spindle events.
    - sfreq (float): Sampling frequency.
    - window (float): Time window (default: ±1.5s around SO trough).

    Returns:
    - pd.DataFrame: Coupled SO-Spindle events.
    """
    stage_mapping = {0: "Wake", 1: "N1", 2: "N2", 3: "SWS", 4: "REM"}
    sfreq = raw.info['sfreq']
    merged = SO_candidates.merge(SP_candidates, on=['ch_name', 'stage'])
    filtered = merged.query("peak_time >= trough_time - @window and peak_time <= trough_time + @window").copy()
    filtered['Time_diff'] = filtered['peak_time'] - filtered['trough_time']
    filtered['Abs_Time_diff'] = filtered['Time_diff'].abs()

    total_spindles = len(SP_candidates)
    total_SOs = len(SO_candidates)
    total_couplings = len(filtered)

    cp_percent = []
    SPcSO_all = (total_couplings / total_spindles) if total_spindles > 0 else 0
    SOcSP_all = (total_couplings / total_SOs) if total_SOs > 0 else 0
    cp_percent.append({'stage': 'all', 'SPcSO': SPcSO_all, 'SOcSP': SOcSP_all})

    for stage in filtered['stage'].unique():
        total_stage_spindles = len(SP_candidates[SP_candidates['stage'] == stage])
        total_stage_SOs = len(SO_candidates[SO_candidates['stage'] == stage])
        total_stage_couplings = len(filtered[filtered['stage'] == stage])

        SPcSO_stage = (total_stage_couplings / total_stage_spindles) if total_stage_spindles > 0 else 0
        SOcSP_stage = (total_stage_couplings / total_stage_SOs) if total_stage_SOs > 0 else 0

        stage_name = stage_mapping.get(stage, f"Unknown ({stage})")
        cp_percent.append({'stage': stage, 'SPcSO': SPcSO_stage, 'SOcSP': SOcSP_stage})
        if verbose:
            print(f"Spindles coupled with SOs in stage \033[1m{stage_name}\033[0m: {SPcSO_stage * 100:.2f}%")
            print(f"SOs coupled with spindles in stage \033[1m{stage_name}\033[0m: {SOcSP_stage * 100:.2f}%")

    cp_percent_df = pd.DataFrame(cp_percent)
    cp_percent_df['channel'] = 'all'

    cp_percent_channel_list = []
    # Group the filtered events by stage and channel
    for (stage, ch), group in filtered.groupby(['stage', 'ch_name']):
        couplings = group.shape[0]
        sp_channel_count = SP_candidates[(SP_candidates['stage'] == stage) & (SP_candidates['ch_name'] == ch)].shape[0]
        so_channel_count = SO_candidates[(SO_candidates['stage'] == stage) & (SO_candidates['ch_name'] == ch)].shape[0]
        SPcSO_channel = (couplings / sp_channel_count) if sp_channel_count > 0 else 0
        SOcSP_channel = (couplings / so_channel_count) if so_channel_count > 0 else 0
        cp_percent_channel_list.append({
            'stage': stage,
            'ch_name': ch,
            'SPcSO': SPcSO_channel,
            'SOcSP': SOcSP_channel
        })
    cp_percent_channel_df = pd.DataFrame(cp_percent_channel_list)
    # Rename 'ch_name' to 'channel' for consistency
    cp_percent_channel_df.rename(columns={'ch_name': 'channel'}, inplace=True)

    # Add channel-level coupling metrics for stage 'all'
    cp_percent_all_channel_list = []
    # Use filtered coupling counts for all channels regardless of stage
    for ch, couplings in filtered['ch_name'].value_counts().items():
        sp_channel_count = SP_candidates[SP_candidates['ch_name'] == ch].shape[0]
        so_channel_count = SO_candidates[SO_candidates['ch_name'] == ch].shape[0]
        SPcSO_channel = (couplings / sp_channel_count) if sp_channel_count > 0 else 0
        SOcSP_channel = (couplings / so_channel_count) if so_channel_count > 0 else 0
        cp_percent_all_channel_list.append({
            'stage': 'all',
            'channel': ch,
            'SPcSO': SPcSO_channel,
            'SOcSP': SOcSP_channel
        })
    cp_percent_all_channel_df = pd.DataFrame(cp_percent_all_channel_list)

    cp_percent_all = pd.concat([cp_percent_df, cp_percent_channel_df, cp_percent_all_channel_df], ignore_index=True)
    print('event_summary\n', event_summary)
    print('cp_percent_all\n', cp_percent_all)
    event_summary = pd.merge(event_summary, cp_percent_all, left_on=['stage', 'channel_y'],
                             right_on=['stage', 'channel'], how='outer')

    # ** Per-Channel Calculation using the same method **
    channel_counts = filtered['ch_name'].value_counts()
    total_spindles_per_channel = SP_candidates['ch_name'].value_counts()
    total_SOs_per_channel = SO_candidates['ch_name'].value_counts()

    spindle_ratio_per_channel = (channel_counts / total_spindles_per_channel * 100).reindex(
        raw.info['ch_names']).fillna(0)
    SO_ratio_per_channel = (channel_counts / total_SOs_per_channel * 100).reindex(raw.info['ch_names']).fillna(0)

    SO_measure = pd.DataFrame({
        'ch_name': raw.info['ch_names'],
        'Spindles_Coupled_with_SOs': spindle_ratio_per_channel.values
    })

    # Generate topomap
    picks = mne.pick_types(raw.info, eeg=True, exclude='bads')
    SO_measure = SO_measure.set_index('ch_name').reindex([raw.ch_names[pick] for pick in picks]).reset_index()

    if verbose:
        fig, ax = plt.subplots(figsize=(4.5, 4.5))
        im, _ = mne.viz.plot_topomap(SO_measure['Spindles_Coupled_with_SOs'].values, raw.info, axes=ax, show=False,
                                     cmap="Blues")

        im.set_clim(vmin=0, vmax=SO_measure['Spindles_Coupled_with_SOs'].max())
        plt.colorbar(im, ax=ax)
        ax.set_title('Spindles Coupled with SOs (%)')
        plt.show()

    return filtered[['ch_name', 'stage', 'trough_time', 'peak_time', 'amplitude', 'Time_diff']].rename(
        columns={'trough_time': 'SO_trough_time', 'peak_time': 'SP_peak_time',
                 'amplitude': 'SP_amplitude'}), event_summary



import mne
import numpy as np
import pandas as pd
from scipy.io import loadmat

from somnopy.utils import good_epoch_dur


def load_raw(file_path):
    if file_path.endswith('.vhdr'):
        return mne.io.read_raw_brainvision(file_path, preload=True, verbose='ERROR')
    elif file_path.endswith('.edf'):
        tbd = mne.io.read_raw_edf(file_path, preload=True, verbose='ERROR')
        return tbd
    elif file_path.endswith('.fif'):
        return mne.io.read_raw_fif(file_path, preload=True, verbose='ERROR')
    elif file_path.endswith('.set'):
        return mne.io.read_raw_eeglab(file_path, preload=True, verbose='ERROR')
    elif file_path.endswith('.bdf'):
        return mne.io.read_raw_bdf(file_path, preload=True, verbose='ERROR')
    elif file_path.endswith('.cnt'):
        return mne.io.read_raw_cnt(file_path, preload=True, verbose='ERROR')
    else:
        raise ValueError(f"Unsupported file format: {file_path}")


def load_hypnogram_data(file_path):
    """
    Load hypnogram data for a given subject.

    Parameters
    ----------
    file_path : str or None
        File path to the scoring file. If None, a default path is constructed.

    Returns
    -------
    numpy.ndarray
        Array of sleep stage values.
    """
    try:
        return load_hypnogram_data_mat(file_path)
    except:
        return load_hypnogram_data_txt(file_path)


def load_hypnogram_data_mat(file_path):
    """
    Load hypnogram data for a given subject.

    Parameters
    ----------
    file_path : str or None
        File path to the scoring file. If None, a default path is constructed.

    Returns
    -------
    numpy.ndarray
        Array of sleep stage values.
    """
    scoring = loadmat(file_path)
    stages = scoring['stageData']['stages']
    stages = np.concatenate([np.concatenate(stage) for stage in stages])
    stages_df = pd.DataFrame(stages, columns=['stages'])
    stages[stages == 5] = '4'
    stages[stages == 7] = '0'
    stages_df.iloc[-1] = 0
    stages_df['stages'] = stages_df['stages'].astype(int)

    return stages_df['stages'].values.flatten()


def load_hypnogram_data_txt(file_path):
    """
    Load hypnogram data for a given subject.

    Parameters
    ----------
    file_path : str or None
        File path to the scoring file. If None, a default path is constructed.

    Returns
    -------
    numpy.ndarray
        Array of sleep stage values.
    """
    scoring = np.genfromtxt(file_path, skip_header=True, dtype=None)
    stages = [stage[0] for stage in scoring]
    stages_df = pd.DataFrame(stages, columns=['stages'])
    stages[stages == 5] = '4'
    stages[stages == 7] = '0'
    stages_df.iloc[-1] = 0
    stages_df['stages'] = stages_df['stages'].astype(int)

    return stages_df['stages'].values.flatten()


def hypnogram_segment(raw, interval=30, bad_epoch=True, file_paths=None):
    """
    Segment hypnogram data into intervals corresponding to sleep stages.

    Parameters
    ----------
    raw: mne.io.Raw
        An instance of Raw EEG data from MNE Python.
    interval : int, optional
        Interval duration in seconds. Default is 30.
    bad_epoch : bool, optional
        Whether to mark stage 6 as bad epochs. Default is True.
    file_paths : str or None
        File path to the scoring file.

    Returns
    -------
    list of tuples
        List of tuples containing (stage, segment length, valid duration).
    """
    if not file_paths:
        print('Please provide a file path for your scoring file.')
    hypno = load_hypnogram_data(file_path=file_paths)
    stage_segments = []
    cur_stage, cnt, seg_start = hypno[0], 0, 0

    for i, stage in enumerate(hypno):
        if stage == cur_stage:
            cnt += 1
        else:
            seg_len = cnt * interval
            seg_end = seg_start + seg_len
            valid_dur = good_epoch_dur(raw, seg_start, seg_end, seg_len)
            stage_segments.append((cur_stage, seg_len, valid_dur))
            cur_stage, cnt, seg_start = stage, 1, seg_end

        if bad_epoch:
            if stage == 6:
                onset = i * interval
                bad_epoch = mne.Annotations(
                    onset=onset,
                    duration=interval,
                    description='Bad_epoch',
                    orig_time=None
                )
                raw.set_annotations(bad_epoch)

    seg_len = cnt * interval
    seg_end = min(seg_start + seg_len, raw.times[-1])  # Ensure segment end is not later than raw_end_time
    seg_len = seg_end - seg_start  # Recalculate the segment length if needed
    valid_dur = good_epoch_dur(raw, seg_start, seg_end, seg_len)
    stage_segments.append((cur_stage, seg_len, valid_dur))

    return stage_segments



import matplotlib.pyplot as plt
import mne
import numpy as np
import pandas as pd
from mpl_toolkits.axes_grid1.inset_locator import inset_axes
from scipy.signal import hilbert, butter, filtfilt
from scipy.stats import zscore, t, ttest_rel, circmean


def plot_SO(SO_candidate, raw, grp_thres=0.5, sync_trough=True, interact=True, multilayer=False):
    from mpl_toolkits.axes_grid1.inset_locator import inset_axes

    # Pick only EEG channels from raw
    picks = mne.pick_types(raw.info, eeg=True, exclude='bads')
    eeg_data = raw.get_data(picks=picks)

    # Get the sampling frequency (sfreq)
    sfreq = raw.info['sfreq']

    # Create a new Info object containing only the EEG channels
    eeg_info = mne.pick_info(raw.info, picks)

    # Apply Butterworth zero-phase filter (0.5-1.25 Hz) to the EEG data
    nyquist = sfreq / 2
    low = 0.5 / nyquist
    high = 1.25 / nyquist
    b, a = butter(2, [low, high], btype='bandpass')
    eeg_data = filtfilt(b, a, eeg_data, axis=1)

    # Create a new column for grouping SOs with None values initially
    SO_candidate['grp_SO'] = None

    # Initialize a unique group counter
    group_id = 1

    # Sort the DataFrame by trough_time for easier sequential comparison
    SO_candidate = SO_candidate.sort_values(by='trough_time').reset_index(drop=True)

    # Loop through each row and assign group IDs
    for idx, row in SO_candidate.iterrows():
        if pd.notnull(SO_candidate.at[idx, 'grp_SO']):
            continue

        SO_candidate.at[idx, 'grp_SO'] = group_id
        group_trough_times = [row['trough_time']]

        while True:
            candidates = (SO_candidate['trough_time'] - np.mean(group_trough_times)).abs() <= grp_thres
            candidates = candidates & (SO_candidate['ch_id'] != row['ch_id']) & SO_candidate['grp_SO'].isnull()

            candidate_indices = SO_candidate.index[candidates].tolist()
            if not candidate_indices:
                break

            for candidate_idx in candidate_indices:
                SO_candidate.at[candidate_idx, 'grp_SO'] = group_id
                group_trough_times.append(SO_candidate.at[candidate_idx, 'trough_time'])

        group_id += 1

    grouped = SO_candidate.groupby('grp_SO')

    if interact:
        plt.ion()
    else:
        plt.ioff()

    for group_id, group in grouped:
        if len(group) <= 1:
            continue

        fig, ax = plt.subplots(figsize=(10, 6))

        channel_indices = []
        all_signals = []
        avg_trough_time = group['trough_time'].mean()
        start_time = avg_trough_time - 1
        end_time = avg_trough_time + 1
        start_idx = int(start_time * sfreq)
        end_idx = int(end_time * sfreq)

        amplitude_scaling = 10 / np.sqrt(len(group))
        trough_points = []

        for _, row in group.iterrows():
            ch_idx = row['ch_id'] - 1
            trough_time = row['trough_time']
            channel_indices.append(ch_idx)

            signal = zscore(eeg_data[ch_idx, start_idx:end_idx])
            time = np.arange(start_idx, end_idx) / sfreq
            all_signals.append(signal)

            # Plot the signal in green
            ax.plot(time, signal * amplitude_scaling + (len(channel_indices) - 1) * 7, color='green', linewidth=2)

            # Mark the trough point
            trough_idx = int(trough_time * sfreq) - start_idx
            ax.plot(time[trough_idx], signal[trough_idx] * amplitude_scaling + (len(channel_indices) - 1) * 7, 'ko')
            trough_points.append(
                (time[trough_idx], signal[trough_idx] * amplitude_scaling + (len(channel_indices) - 1) * 7))

        # Connect trough points with green lines
        for i in range(len(trough_points) - 1):
            ax.plot([trough_points[i][0], trough_points[i + 1][0]],
                    [trough_points[i][1], trough_points[i + 1][1]], 'g--', linewidth=1)

        ax.axvline(avg_trough_time, color='red', linestyle='--')
        ax.set_yticks([(i - 1) * 7 for i in range(1, len(channel_indices) + 1)])
        ax.set_yticklabels([raw.info['ch_names'][ch_idx] for ch_idx in channel_indices])
        ax.set_xlabel('Time (s)')
        ax.set_ylabel('Channels')

        # Topomap in the upper-right corner
        ax_inset_topomap = inset_axes(ax, width="20%", height="20%", loc='upper right')
        ptp_amplitudes = []
        for ch_idx in range(eeg_data.shape[0]):
            ptp_signal = eeg_data[ch_idx, start_idx:end_idx]
            ptp_amplitude = np.ptp(ptp_signal)
            ptp_amplitudes.append(ptp_amplitude)

        mask = np.zeros(len(eeg_info['ch_names']), dtype=bool)
        mask[channel_indices] = True

        mne.viz.plot_topomap(ptp_amplitudes, eeg_info, axes=ax_inset_topomap, show=False, mask=mask,
                             mask_params=dict(markersize=8, markerfacecolor='r', alpha=0.5), outlines='head',
                             cmap='viridis')

        if len(all_signals) > 1:
            lags = []
            correlations = []
            for i in range(len(all_signals)):
                for j in range(i + 1, len(all_signals)):
                    signal_i = all_signals[i]
                    signal_j = all_signals[j]

                    max_lag_samples = int(0.2 * sfreq)
                    corr = np.correlate(signal_i, signal_j, mode='same')
                    auto_corr_i = np.correlate(signal_i, signal_i, mode='same')[len(signal_i) // 2]
                    auto_corr_j = np.correlate(signal_j, signal_j, mode='same')[len(signal_j) // 2]
                    norm_corr = corr / np.sqrt(auto_corr_i * auto_corr_j)

                    valid_corr = norm_corr[
                                 len(signal_i) // 2 - max_lag_samples: len(signal_i) // 2 + max_lag_samples + 1]
                    max_corr_idx = np.argmax(valid_corr)
                    max_corr = valid_corr[max_corr_idx]
                    lag_ms = (max_corr_idx - max_lag_samples) / sfreq * 1000

                    correlations.append(max_corr)
                    lags.append(lag_ms)

            # Time lag histogram (upper left, square)
            ax_inset_lag = inset_axes(ax, width="10%", height="15%", loc='lower left')
            ax_inset_lag.hist(lags, bins=20, color='green', alpha=0.7)
            ax_inset_lag.set_xlim(-200, 200)
            ax_inset_lag.set_xticks([-200, 0, 200])
            ax_inset_lag.set_yticks([])
            ax_inset_lag.set_title('Lag', fontsize=8)

        if interact:
            plt.pause(0.2)
        else:
            plt.show()


def plot_SP(SP_candidate, raw, grp_thres=0.5, sync_peak=False, interact=False, multilayer=False):
    # Pick only EEG channels from raw
    picks = mne.pick_types(raw.info, eeg=True, exclude='bads')
    eeg_data = raw.get_data(picks=picks)

    # Get the sampling frequency (sfreq)
    sfreq = raw.info['sfreq']

    # Create a new Info object containing only the EEG channels
    eeg_info = mne.pick_info(raw.info, picks)

    # Apply Butterworth zero-phase filter (12-16 Hz) to the EEG data
    nyquist = sfreq / 2
    low = 12 / nyquist
    high = 16 / nyquist
    b, a = butter(2, [low, high], btype='bandpass')
    eeg_data = filtfilt(b, a, eeg_data, axis=1)

    # Create a new column for grouping SPs with None values initially
    SP_candidate['grp_SP'] = None

    # Initialize a unique group counter
    group_id = 1

    # Sort the DataFrame by peak_time for easier sequential comparison
    SP_candidate = SP_candidate.sort_values(by='peak_time').reset_index(drop=True)

    # Loop through each row and assign group IDs
    for idx, row in SP_candidate.iterrows():
        if pd.notnull(SP_candidate.at[idx, 'grp_SP']):
            continue

        SP_candidate.at[idx, 'grp_SP'] = group_id
        group_peak_times = [row['peak_time']]

        while True:
            candidates = (SP_candidate['peak_time'] - np.mean(group_peak_times)).abs() <= grp_thres
            candidates = candidates & (SP_candidate['ch_id'] != row['ch_id']) & SP_candidate['grp_SP'].isnull()

            candidate_indices = SP_candidate.index[candidates].tolist()
            if not candidate_indices:
                break

            for candidate_idx in candidate_indices:
                SP_candidate.at[candidate_idx, 'grp_SP'] = group_id
                group_peak_times.append(SP_candidate.at[candidate_idx, 'peak_time'])

        group_id += 1

    grouped = SP_candidate.groupby('grp_SP')

    if interact:
        plt.ion()
    else:
        plt.ioff()

    for group_id, group in grouped:
        if len(group) <= 1:
            continue

        fig, ax = plt.subplots(figsize=(10, 6))

        channel_indices = []
        all_signals = []
        avg_peak_time = group['peak_time'].mean()
        start_time = avg_peak_time - 1
        end_time = avg_peak_time + 1
        start_idx = int(start_time * sfreq)
        end_idx = int(end_time * sfreq)

        amplitude_scaling = 10 / np.sqrt(len(group))
        peak_points = []

        for _, row in group.iterrows():
            ch_idx = row['ch_id'] - 1
            peak_time = row['peak_time']
            channel_indices.append(ch_idx)

            signal = zscore(eeg_data[ch_idx, start_idx:end_idx])
            time = np.arange(start_idx, end_idx) / sfreq
            all_signals.append(signal)

            # Plot the signal in green
            ax.plot(time, signal * amplitude_scaling + (len(channel_indices) - 1) * 7, color='green', linewidth=2)

            # Mark the peak point
            peak_idx = int(peak_time * sfreq) - start_idx
            ax.plot(time[peak_idx], signal[peak_idx] * amplitude_scaling + (len(channel_indices) - 1) * 7, 'ko')
            peak_points.append((time[peak_idx], signal[peak_idx] * amplitude_scaling + (len(channel_indices) - 1) * 7))

        # Connect peak points with green lines
        for i in range(len(peak_points) - 1):
            ax.plot([peak_points[i][0], peak_points[i + 1][0]],
                    [peak_points[i][1], peak_points[i + 1][1]], 'g--', linewidth=1)

        ax.axvline(avg_peak_time, color='red', linestyle='--')
        ax.set_yticks([(i - 1) * 7 for i in range(1, len(channel_indices) + 1)])
        ax.set_yticklabels([raw.info['ch_names'][ch_idx] for ch_idx in channel_indices])
        ax.set_xlabel('Time (s)')
        ax.set_ylabel('Channels')

        # Topomap in the upper-right corner
        ax_inset_topomap = inset_axes(ax, width="20%", height="20%", loc='upper right')
        ptp_amplitudes = []
        for ch_idx in range(eeg_data.shape[0]):
            ptp_signal = eeg_data[ch_idx, start_idx:end_idx]
            ptp_amplitude = np.ptp(ptp_signal)
            ptp_amplitudes.append(ptp_amplitude)

        mask = np.zeros(len(eeg_info['ch_names']), dtype=bool)
        mask[channel_indices] = True

        mne.viz.plot_topomap(ptp_amplitudes, eeg_info, axes=ax_inset_topomap, show=False, mask=mask,
                             mask_params=dict(markersize=8, markerfacecolor='r', alpha=0.5), outlines='head',
                             cmap='viridis')

        if len(all_signals) > 1:
            lags = []
            correlations = []
            for i in range(len(all_signals)):
                for j in range(i + 1, len(all_signals)):
                    signal_i = all_signals[i]
                    signal_j = all_signals[j]

                    max_lag_samples = int(0.2 * sfreq)
                    corr = np.correlate(signal_i, signal_j, mode='same')
                    auto_corr_i = np.correlate(signal_i, signal_i, mode='same')[len(signal_i) // 2]
                    auto_corr_j = np.correlate(signal_j, signal_j, mode='same')[len(signal_j) // 2]
                    norm_corr = corr / np.sqrt(auto_corr_i * auto_corr_j)

                    valid_corr = norm_corr[
                                 len(signal_i) // 2 - max_lag_samples: len(signal_i) // 2 + max_lag_samples + 1]
                    max_corr_idx = np.argmax(valid_corr)
                    max_corr = valid_corr[max_corr_idx]
                    lag_ms = (max_corr_idx - max_lag_samples) / sfreq * 1000

                    correlations.append(max_corr)
                    lags.append(lag_ms)

            # Time lag histogram (upper left, square)
            ax_inset_lag = inset_axes(ax, width="10%", height="15%", loc='lower left')
            ax_inset_lag.hist(lags, bins=20, color='green', alpha=0.7)
            ax_inset_lag.set_xlim(-200, 200)
            ax_inset_lag.set_xticks([-200, 0, 200])
            ax_inset_lag.set_yticks([])
            ax_inset_lag.set_title('Lag', fontsize=8)

        if interact:
            plt.pause(0.2)
        else:
            plt.show()


def compute_pac_metrics(events, raw, window_size, eeg_data_dict, n_bins, sfreq, stage_dur):
    """
    Compute PAC metrics for a given set of events (subset for a channel and optionally a stage).
    Returns a dictionary with metrics.
    """
    peaks_so_phase = []
    valid_events_count = 0
    for _, event in events.iterrows():
        ch = event['ch_name']
        so_trough_time = event['SO_trough_time']
        sp_peak_time = event['SP_peak_time']
        trough_idx = int(so_trough_time * sfreq)
        sp_idx = int(sp_peak_time * sfreq)
        if trough_idx - window_size < 0 or trough_idx + window_size >= raw.n_times:
            continue
        local_window = eeg_data_dict[ch][trough_idx - window_size: trough_idx + window_size]
        analytic_signal = hilbert(local_window)
        phase_data = np.angle(analytic_signal)
        relative_spindle_peak_idx = sp_idx - (trough_idx - window_size)
        if 0 <= relative_spindle_peak_idx < len(phase_data):
            peaks_so_phase.append(phase_data[relative_spindle_peak_idx])
            valid_events_count += 1
    peaks_so_phase = np.array(peaks_so_phase)
    if len(peaks_so_phase) > 0:
        pp = np.angle(np.exp(1j * circmean(peaks_so_phase)))
        amp_hist, _ = np.histogram(peaks_so_phase, bins=np.linspace(-np.pi, np.pi, n_bins + 1), density=True)
        U = np.ones(n_bins) / n_bins
        P = amp_hist / np.sum(amp_hist)
        D_KL = np.nansum(P * np.log(P / U))
        mi = D_KL / np.log(n_bins) if np.log(n_bins) > 0 else 0
        mvl = np.abs(np.mean(np.exp(1j * peaks_so_phase)))
        R = mvl
        z = len(peaks_so_phase) * R ** 2
        p_value = np.exp(-z) * (1 + z)
    else:
        pp = np.nan
        mi = np.nan
        mvl = np.nan
        z = np.nan
        p_value = np.nan
    # Compute coupling density: for a single channel, use (number of events / stage duration) * 30.
    cp_count = events.shape[0]
    cp_density = (cp_count / stage_dur) * 30 if stage_dur > 0 else 0
    return {
        'preferred_phase': pp,
        'modulation_index': mi,
        'mean_vector_length': mvl,
        'rayleigh_z': z,
        'coupling_density': cp_density
    }


def pac(raw, coupled_events, event_summary, verbose=True):
    """
    Compute Peri-Event Time Histogram (PETH) and Phase-Amplitude Coupling (PAC)
    between SO and spindle events in a single function.

    Parameters:
    - coupled_events (pd.DataFrame): DataFrame containing SO-Spindle coupling events.
    - raw (mne.io.Raw): Raw EEG data.

    Prints:
    - Preferred Phase, Modulation Index, Mean Vector Length, Rayleigh z, and p-value.

    Plots:
    - Left: PETH histogram.
    - Right: PAC polar plot.
    """
    sfreq = raw.info['sfreq']  # Sampling frequency
    window = 1.2  # Time window for PETH
    bins = 20  # Bins for PETH histogram
    window_size = int(window * sfreq)  # Convert ±1.2s window into samples
    time_diffs = coupled_events['Time_diff'].dropna().values  # Remove NaN values

    # **Compute histogram (normalized)**
    peth_bin_edges = np.linspace(-window, window, bins + 1)

    # **Extract SO waveform around each SO trough (±1.2s)**
    time_range = np.arange(-window, window, 1 / sfreq)
    avg_so_wave = np.zeros_like(time_range)

    num_valid_events = 0
    SO_waveform = []
    for _, event in coupled_events.iterrows():
        ch_name = event['ch_name']
        so_trough_time = event['SO_trough_time']

        trough_idx = int(so_trough_time * sfreq)
        start_idx = trough_idx - window_size
        end_idx = trough_idx + window_size

        if start_idx >= 0 and end_idx < raw.n_times:
            eeg_data = raw.get_data(picks=ch_name)[0]
            avg_so_wave += eeg_data[start_idx:end_idx]
            num_valid_events += 1

    if num_valid_events > 0:
        avg_so_wave /= num_valid_events  # Compute mean SO waveform

    avg_so_wave *= 1e6

    SO_waveform.append({'stage': 'all', 'so_waveform': ','.join(map(str, avg_so_wave))})

    # **PAC Computation**
    unique_channels = coupled_events['ch_name'].unique()
    eeg_data_dict = {ch: raw.get_data(picks=ch)[0] for ch in unique_channels}

    so_trough_indices = (coupled_events['SO_trough_time'] * sfreq).astype(int)
    spindle_peak_indices = (coupled_events['SP_peak_time'] * sfreq).astype(int)

    valid_mask = (so_trough_indices >= window_size) & (so_trough_indices < raw.n_times - window_size)

    peaks_so_phase_all = []

    for ch, so_idx, sp_idx in zip(coupled_events['ch_name'][valid_mask],
                                  so_trough_indices[valid_mask],
                                  spindle_peak_indices[valid_mask]):
        local_window = eeg_data_dict[ch][so_idx - window_size: so_idx + window_size]
        analytic_signal = hilbert(local_window)
        phase_data = np.angle(analytic_signal)

        relative_spindle_peak_idx = sp_idx - (so_idx - window_size)
        if 0 <= relative_spindle_peak_idx < len(phase_data):
            peaks_so_phase_all.append(phase_data[relative_spindle_peak_idx])

    peaks_so_phase_all = np.array(peaks_so_phase_all)

    if len(peaks_so_phase_all) == 0:
        print("No valid SO phases found.")
        return

    # Always calculate both all-stage data and individual-stage data
    stage_mapping = {0: "Wake", 1: "N1", 2: "N2", 3: "SWS", 4: "REM"}
    coupling_results = []

    # Compute all-stage metrics
    peaks_so_phase_all = np.array(peaks_so_phase_all)
    n_bins = 20
    pac_bin_edges = np.linspace(-np.pi, np.pi, n_bins + 1)
    U = np.ones(n_bins) / n_bins

    if len(peaks_so_phase_all) > 0:
        pp_all = np.angle(np.exp(1j * circmean(peaks_so_phase_all)))
        amp_hist_all, _ = np.histogram(peaks_so_phase_all, bins=np.linspace(-np.pi, np.pi, n_bins + 1), density=True)
        P_all = amp_hist_all / np.sum(amp_hist_all)
        D_KL_all = np.nansum(P_all * np.log(P_all / U))
        mi_all = D_KL_all / np.log(n_bins) if np.log(n_bins) > 0 else 0
        mvl_all = np.abs(np.mean(np.exp(1j * peaks_so_phase_all)))
        R_all = np.abs(np.mean(np.exp(1j * peaks_so_phase_all)))
        z_all = len(peaks_so_phase_all) * R_all ** 2
        p_all = np.exp(-z_all) * (1 + z_all)

        cp_count_all = coupled_events.shape[0]
        total_stage_dur = event_summary.loc[event_summary['stage'] == 'all', 'stage_dur'].values[0]
        num_channels_all = coupled_events['ch_name'].nunique()
        cp_density_all = (cp_count_all / (total_stage_dur * num_channels_all)) * 30 if total_stage_dur > 0 else 0

        coupling_results.append({
            'stage': 'all',
            'preferred_phase': pp_all,
            'modulation_index': mi_all,
            'mean_vector_length': mvl_all,
            'rayleigh_z': z_all,
            'coupling_density': cp_density_all
        })

    # Compute individual-stage metrics
    for stage in coupled_events['stage'].unique():
        stage_name = stage_mapping.get(stage, f"Unknown ({stage})")
        stage_data = coupled_events[coupled_events['stage'] == stage]
        peaks_so_phase_stage = []
        avg_so_wave_stage = np.zeros_like(time_range)
        num_valid_events_stage = 0

        so_trough_indices_stage = (stage_data['SO_trough_time'] * sfreq).astype(int)
        spindle_peak_indices_stage = (stage_data['SP_peak_time'] * sfreq).astype(int)

        valid_mask_stage = (so_trough_indices_stage >= window_size) & (
                so_trough_indices_stage < raw.n_times - window_size)

        for ch, so_idx, sp_idx in zip(stage_data['ch_name'][valid_mask_stage],
                                      so_trough_indices_stage[valid_mask_stage],
                                      spindle_peak_indices_stage[valid_mask_stage]):
            local_window = eeg_data_dict[ch][so_idx - window_size: so_idx + window_size]
            analytic_signal = hilbert(local_window)
            phase_data = np.angle(analytic_signal)

            relative_spindle_peak_idx = sp_idx - (so_idx - window_size)
            if 0 <= relative_spindle_peak_idx < len(phase_data):
                peaks_so_phase_stage.append(phase_data[relative_spindle_peak_idx])

        peaks_so_phase_stage = np.array(peaks_so_phase_stage)

        if len(peaks_so_phase_stage) > 0:
            pp_stage = np.angle(np.exp(1j * circmean(peaks_so_phase_stage)))
            amp_hist_stage, _ = np.histogram(peaks_so_phase_stage, bins=np.linspace(-np.pi, np.pi, n_bins + 1),
                                             density=True)
            P_stage = amp_hist_stage / np.sum(amp_hist_stage)
            D_KL_stage = np.nansum(P_stage * np.log(P_stage / U))
            mi_stage = D_KL_stage / np.log(n_bins) if np.log(n_bins) > 0 else 0
            mvl_stage = np.abs(np.mean(np.exp(1j * peaks_so_phase_stage)))
            R_stage = np.abs(np.mean(np.exp(1j * peaks_so_phase_stage)))
            z_stage = len(peaks_so_phase_stage) * R_stage ** 2
            p_stage = np.exp(-z_stage) * (1 + z_stage)

            cp_count = coupled_events.groupby('stage').size()
            stage_dur = event_summary.loc[event_summary['stage'] == stage, 'stage_dur'].values[0]
            num_cp_event = cp_count.get(stage, 0)
            num_channels_stage = stage_data['ch_name'].nunique()
            cp_density_stage = (num_cp_event / (stage_dur * num_channels_stage)) * 30 if stage_dur > 0 else 0

            coupling_results.append({
                'stage': stage,
                'preferred_phase': pp_stage,
                'modulation_index': mi_stage,
                'mean_vector_length': mvl_stage,
                'rayleigh_z': z_stage,
                'coupling_density': cp_density_stage
            })

            if verbose:
                print(f"\033[1mStage {stage_name}:\033[0m")
                print(f"Preferred Phase:       {round(pp_stage, 3)} rad")
                print(f"Modulation Index:      {mi_stage:.3f}")
                print(f"Mean Vector Length:    {mvl_stage:.3f}")
                print(f"Rayleigh z:            {z_stage:.3f}")
                print(f"p-value:               {p_stage:.3f}")
                print(f"Coupling Density:      {cp_density_stage:.3f}")

        for _, event in stage_data.iterrows():
            ch_name = event['ch_name']
            so_trough_time = event['SO_trough_time']

            trough_idx = int(so_trough_time * sfreq)
            start_idx = trough_idx - window_size
            end_idx = trough_idx + window_size

            if start_idx >= 0 and end_idx < raw.n_times:
                eeg_data = raw.get_data(picks=ch_name)[0]
                avg_so_wave_stage += eeg_data[start_idx:end_idx]
                num_valid_events_stage += 1

        if num_valid_events_stage > 0:
            avg_so_wave_stage /= num_valid_events_stage
        avg_so_wave_stage *= 1e6

        SO_waveform.append({'stage': stage, 'so_waveform': ','.join(map(str, avg_so_wave_stage))})

    # Merge results into event_summary
    coupling_df = pd.DataFrame(coupling_results)
    coupling_df['channel'] = 'all'

    per_channel_results = []
    # Prepare a dictionary for stage durations from event_summary input
    stage_dur_dict = {row['stage']: row['stage_dur'] for _, row in event_summary.iterrows()}
    # Process each channel for aggregated (all stages) metrics
    for ch in coupled_events['ch_name'].unique():
        events_ch_all = coupled_events[coupled_events['ch_name'] == ch]
        if events_ch_all.empty:
            continue
        # Get stage duration for 'all'
        total_stage_dur = stage_dur_dict.get('all', np.nan)
        metrics_all = compute_pac_metrics(events_ch_all, raw, window_size, eeg_data_dict, n_bins, sfreq,
                                          total_stage_dur)
        metrics_all.update({'stage': 'all', 'channel': ch})
        per_channel_results.append(metrics_all)
        # Process each stage for this channel
        for stage in events_ch_all['stage'].unique():
            events_ch_stage = events_ch_all[events_ch_all['stage'] == stage]
            stage_dur = stage_dur_dict.get(stage, np.nan)
            metrics_stage = compute_pac_metrics(events_ch_stage, raw, window_size, eeg_data_dict, n_bins, sfreq,
                                                stage_dur)
            metrics_stage.update({'stage': stage, 'channel': ch})
            per_channel_results.append(metrics_stage)
    per_channel_coupling_df = pd.DataFrame(per_channel_results)
    coupling_all = pd.concat([coupling_df, per_channel_coupling_df], ignore_index=True)
    event_summary = pd.merge(event_summary, coupling_all, on=['stage', 'channel'], how='outer')

    if verbose:
        # **Plot PETH + PAC**
        fig = plt.figure(figsize=(8, 3.6))
        gs = plt.GridSpec(1, 2, width_ratios=[1.9, 1])

        # Left: PETH
        ax1 = fig.add_subplot(gs[0, 0])
        ax1.hist(time_diffs, bins=peth_bin_edges, facecolor='none', edgecolor='#4682B4', linewidth=2, alpha=1,
                 density=True, label='PETH')
        ax1.axvline(0, color='black', lw=2, linestyle='--', label='SO Trough')
        ax1.set_xlabel('Time relative to SO Trough (s)')
        ax1.set_ylabel('Density')
        ax1.set_title('Peri-Event Time Histogram (PETH)', pad=10)

        ax2 = ax1.twinx()
        ax2.plot(time_range, avg_so_wave, color='black', linewidth=2, label='Averaged SO Waveform')
        ax2.set_ylim([avg_so_wave.min() * 1.5, avg_so_wave.max() * 1.2])
        ax2.spines['right'].set_visible(False)
        ax2.set_yticklabels([])
        ax2.set_yticks([])

        # Right: PAC polar plot
        theta = (pac_bin_edges[:-1] + pac_bin_edges[1:]) / 2
        r = amp_hist_all

        ax3 = fig.add_subplot(gs[0, 1], projection='polar')
        ax3.bar(theta, r, width=(2 * np.pi / n_bins), facecolor='none', edgecolor='#4682B4', linewidth=2, alpha=1)
        ax3.plot([pp_all, pp_all], [0, max(r)], color='k', lw=2, linestyle='--', label="Preferred Phase")
        ax3.set_yticks([])
        ax3.yaxis.set_visible(False)
        ax3.set_xticks([0, np.pi / 2, np.pi, 3 * np.pi / 2])  # Set only four tick labels
        ax3.set_xticklabels(['0 \n peak', r'$\pi/2$', r'$\pi$', r'$-\pi/2$'])  # Custom labels

        ax3.set_theta_zero_location("E")  # 'East' → Places 0 on the right
        ax3.set_theta_direction(1)  # Counterclockwise direction

        ax3.set_title('Coupling Phase Polar Plot')

        plt.tight_layout()
        plt.show()

    waveform_df = pd.DataFrame(SO_waveform)
    return event_summary, waveform_df


def pac_grp(raw, cp_event_grp, event_summary_grp, SO_waveform, stage='all', chan='all'):
    """
    Compute and plot:
    1. Peri-Event Time Histogram (PETH) with averaged SO waveform, with a permutation test
       to assess the temporal stability of the histogram.
    2. Group-level preferred phase circular plot, where the preferred phase indicator line
       extends from 0 to the mean vector length computed across subjects.

    Parameters:
    -----------
    raw : mne.io.Raw
        Raw EEG data.
    cp_event_grp : pd.DataFrame
        DataFrame containing SO-Spindle coupling events. Must include columns:
        'Time_diff', 'SO_trough_time', 'ch_name', 'subject', and 'stage'.
    event_summary_grp : pd.DataFrame
        DataFrame containing participant-level PAC metrics
    stage : str, optional
        Specify which stage to plot. Enter 'all' to aggregate across all stages, or specify
        a stage such as 'Wake', 'N1', 'N2', 'SWS', or 'REM' to generate a plot for that specific stage.

    Notes:
    ------
    To test the temporal stability of the PETHs, we implemented a permutation test:
      - For each subject, the temporal order of the PETH bins is randomly shuffled 1000 times.
      - The resulting surrogate histograms are averaged for each individual.
      - Paired one-tailed t-tests are performed for each bin comparing the original and surrogate histograms
      - Controled multiple comparisons by a cluster-based permutation test with 5000 permutations.
    """

    stage_mapping = {0: "Wake", 1: "N1", 2: "N2", 3: "SWS", 4: "REM"}
    sfreq = raw.info['sfreq']
    window = 1.2
    bins = 20
    window_size = int(window * sfreq)

    # Determine stages to iterate over
    if stage == 'all':
        stages = ['all']
        cp_event_grp = cp_event_grp.copy()
        event_summary_grp = event_summary_grp[event_summary_grp['stage'] == 'all']
    else:
        inv_stage_mapping = {v: k for k, v in stage_mapping.items()}
        if stage not in inv_stage_mapping:
            raise ValueError("Invalid stage input. Must be one of 'all', 'Wake', 'N1', 'N2', 'SWS', 'REM'.")
        stage_int = inv_stage_mapping[stage]
        stages = [stage_int]

    for stg in stages:
        if stg == 'all':
            stage_name = "Global"
            stage_events = cp_event_grp
            stage_pac_metrics = event_summary_grp
        else:
            stage_name = stage_mapping.get(stg, f"Unknown ({stg})")
            stage_events = cp_event_grp[cp_event_grp['stage'] == stg]
            stage_pac_metrics = event_summary_grp[event_summary_grp['stage'] == stg]

        if chan == 'all':
            stage_pac_metrics = stage_pac_metrics[stage_pac_metrics['channel'] == 'all']
        else:
            if isinstance(chan, list):
                stage_pac_metrics = stage_pac_metrics[stage_pac_metrics['channel'].isin(chan)]
            else:
                stage_pac_metrics = stage_pac_metrics[stage_pac_metrics['channel'] == chan]

            if stage_pac_metrics.empty:
                print(f"Error: Selected channel(s) {chan} not found in event_summary_grp.")
                return

        if stage_events.empty or stage_pac_metrics.empty:
            continue

        time_diffs = stage_events['Time_diff'].dropna().values
        peth_bin_edges = np.linspace(-window, window, bins + 1)
        time_range = np.arange(-window, window, 1 / sfreq)

        # Extract SO waveforms from DataFrame instead of recomputing from raw
        if stg == 'all':
            stage_waveforms = SO_waveform[SO_waveform['stage'] == 'all']['so_waveform']
        else:
            stage_waveforms = SO_waveform[SO_waveform['stage'] == stg]['so_waveform']

        # Convert waveform strings into numerical arrays and average across subjects
        if not stage_waveforms.empty:
            waveform_arrays = np.array([np.fromstring(w, sep=',') for w in stage_waveforms])
            avg_so_wave = waveform_arrays.mean(axis=0)  # Average across subjects
        else:
            avg_so_wave = np.zeros_like(time_range)  # Fallback to zeros if no data

        subjects = stage_events['subject'].unique()
        n_subjects = len(subjects)
        subject_histograms = np.zeros((n_subjects, bins))
        for i, subj in enumerate(subjects):
            subj_data = stage_events[stage_events['subject'] == subj]['Time_diff'].dropna().values
            hist, _ = np.histogram(subj_data, bins=peth_bin_edges, density=True)
            subject_histograms[i, :] = hist
        group_original = subject_histograms.mean(axis=0)

        n_permutations = 1000
        subject_surrogate_avg = np.zeros((n_subjects, bins))
        for i in range(n_subjects):
            surrogate_all = np.zeros((n_permutations, bins))
            for j in range(n_permutations):
                surrogate_all[j, :] = np.random.permutation(subject_histograms[i, :])
            subject_surrogate_avg[i, :] = surrogate_all.mean(axis=0)

        diff = subject_histograms - subject_surrogate_avg
        t_vals, p_vals = ttest_rel(subject_histograms, subject_surrogate_avg, axis=0)

        group_surrogate_avg = subject_surrogate_avg.mean(axis=0)  # Average across subjects

        alpha = 0.05
        t_thresh = t.ppf(1 - alpha, df=n_subjects - 1)

        def find_clusters_t(t_values, t_thresh):
            clusters = []
            cluster_stats = []
            current_cluster = []
            for idx in range(len(t_values)):
                if t_values[idx] > t_thresh:
                    current_cluster.append(idx)
                else:
                    if current_cluster:
                        clusters.append(current_cluster)
                        cluster_stats.append(np.sum(t_values[current_cluster]))
                        current_cluster = []
            if current_cluster:
                clusters.append(current_cluster)
                cluster_stats.append(np.sum(t_values[current_cluster]))
            return clusters, cluster_stats

        observed_clusters, observed_cluster_stats = find_clusters_t(t_vals, t_thresh)

        n_cluster_permutations = 5000
        max_cluster_stats_perm = np.zeros(n_cluster_permutations)

        for perm in range(n_cluster_permutations):
            sign_flip = np.random.choice([1, -1], size=n_subjects)
            diff_perm = diff * sign_flip[:, None]
            t_perm = np.zeros(bins)
            for b in range(bins):
                d = diff_perm[:, b]
                std_d = np.std(d, ddof=1)
                if std_d == 0:
                    t_perm[b] = 0
                else:
                    t_perm[b] = np.mean(d) / (std_d / np.sqrt(n_subjects))
            clusters_perm, cluster_stats_perm = find_clusters_t(t_perm, t_thresh)
            if cluster_stats_perm:
                max_cluster_stats_perm[perm] = np.max(cluster_stats_perm)
            else:
                max_cluster_stats_perm[perm] = 0

        significant_bins = np.zeros(bins, dtype=bool)
        for cluster, cluster_stat in zip(observed_clusters, observed_cluster_stats):
            p_corr = np.mean(max_cluster_stats_perm >= cluster_stat)
            if p_corr < 0.05:
                significant_bins[cluster] = True

        fig = plt.figure(figsize=(8, 3.6))
        gs = plt.GridSpec(1, 2, width_ratios=[1.9, 1])

        ax1 = fig.add_subplot(gs[0, 0])
        ax1.hist(time_diffs, bins=peth_bin_edges, facecolor='none', edgecolor='#4682B4',
                 linewidth=2, alpha=1, density=True, label='PETH')
        ax1.axvline(0, color='black', lw=2, linestyle='--', label='SO Trough')
        bin_centers = (peth_bin_edges[:-1] + peth_bin_edges[1:]) / 2
        ax1.plot(bin_centers, group_surrogate_avg, color='grey', linewidth=2, label='Surrogate Average')
        ax1.set_xlabel('Time relative to SO Trough (s)')
        ax1.set_ylabel('Density')

        ax2 = ax1.twinx()
        ax2.plot(time_range, avg_so_wave, color='black', linewidth=2, label='Averaged SO Waveform')
        ax2.set_ylim([avg_so_wave.min() * 1.5, avg_so_wave.max() * 1.2])
        ax2.spines['right'].set_visible(False)
        ax2.set_yticklabels([])
        ax2.set_yticks([])

        bin_centers = (peth_bin_edges[:-1] + peth_bin_edges[1:]) / 2
        ylim = ax1.get_ylim()
        y_offset = ylim[1] * 1.05
        for i, sig in enumerate(significant_bins):
            if sig:
                ax1.plot(bin_centers[i], y_offset, 'o', color='red')

        ax3 = fig.add_subplot(gs[0, 1], projection='polar')
        pp_data = stage_pac_metrics['preferred_phase'].dropna().values
        if len(pp_data) > 0:
            mean_phase_vector = np.mean(np.exp(1j * pp_data))
            mean_phase = np.angle(mean_phase_vector)
            mvl = np.abs(mean_phase_vector)
            ax3.plot([mean_phase, mean_phase], [0, mvl], color='k', lw=2, label="Mean Preferred Phase")
            ax3.scatter(pp_data, np.ones_like(pp_data), color='#4682B4', s=50, label='Participants', alpha=0.8)

        ax3.set_yticks([])
        ax3.yaxis.set_visible(False)
        ax3.set_xticks([0, np.pi / 2, np.pi, 3 * np.pi / 2])
        ax3.set_xticklabels(['0 \n peak', r'$\pi/2$', r'$\pi$', r'$-\pi/2$'])
        ax3.set_theta_zero_location("E")
        ax3.set_theta_direction(1)

        plt.tight_layout()
        plt.show()



import pytest

from somnopy.preprocessing import load_hypnogram_data


@pytest.fixture
def dummy_data():
    # Setup your dummy EEG data and sampling frequency.
    return {
        'eeg': [0.1, 0.5, 0.3, 0.7],  # Replace with realistic dummy data.
        'sf': 100
    }

def test_detect_spindles(dummy_data):
    result = load_hypnogram_data(dummy_data['eeg'], sf=dummy_data['sf'])
    assert result is not None


import math
import os

import pandas as pd
from mne.io import Raw

from somnopy.event_detection import SO_detection, SP_detection
from somnopy.metrics import pac
from somnopy.preprocessing import load_raw, hypnogram_segment
from somnopy.utils import event_lock, set_up_raw

#TODO:Add stages to SWA
#TODO: Analyze Melissa's data for SWA and Spindles, data in infantro1/data/psg/datatobeanalyzed

#  TODO: add method for saving files on the notebook
#   TODO: Add custom method for spindle detection

# TODO: can compare with Melissa's data

# TODO: Can we look at the spindle frequency


def get_sosp(raw: Raw, stage, interest_stage=('N2', 'SWS'),
             sp_method='Hahn2020',
             so_method='Staresina', coupling=False,
             filter_freq=None, duration=None, filter_type: str = 'fir', l_freq: float = None,
             h_freq: float = None, dur_lower: float = None, dur_upper: float = None,
             baseline: bool = True, verbose: bool = True):
    _, so_candidate, so_summary = SO_detection(raw, stage, target_stage=interest_stage, method=so_method,
                                               baseline=baseline, verbose=verbose, filter_freq=filter_freq,
                                               duration=duration, filter_type=filter_type)
    _, sp_candidate, sp_summary = SP_detection(raw, stage, target_stage=interest_stage, method=sp_method,
                                               l_freq=l_freq, h_freq=h_freq, dur_lower=dur_lower,
                                               dur_upper=dur_upper, baseline=baseline, verbose=verbose)
    event_summary = pd.merge(so_summary, sp_summary, on=['stage'], how='outer')

    cp_event = None
    so_waveform = None

    if coupling:
        cp_event, event_summary = event_lock(raw, so_candidate, sp_candidate, event_summary, verbose=verbose)
        cp_event.insert(0, 'subject', id)
        event_summary, so_waveform = pac(raw, cp_event, event_summary, verbose=verbose)
    return event_summary, cp_event, so_waveform


def get_sosp_for_folder(raw_folder: str, stage_folder: str, interest_stage=('N2', 'SWS'),
                        sp_method='Hahn2020',
                        so_method='Staresina', coupling=True, scoring_dur=30, rerefer=False, chan_limit=None,
                        ch_drop=(),
                        montage_temp="standard_1005", is_montage=True,
                        filter_freq=None, duration=None, filter_type: str = 'fir', l_freq: float = None,
                        h_freq: float = None, dur_lower: float = 0.5, dur_upper: float = math.inf,
                        baseline: bool = True, verbose: bool = True):
    file_paths = [os.path.join(raw_folder, f) for f in os.listdir(raw_folder) if
                  f.endswith(('.vhdr', '.edf', '.fif', '.set', '.fdt', '.bdf', '.cnt'))]
    coupling_event_all = {}
    event_summary_all = {}
    so_waveform_all = {}

    for file_path in file_paths:
        file_name = os.path.splitext(os.path.basename(file_path))[0]
        print(f"\033[1mStart event detection for subject {file_name}.\033[0m")
        raw = load_raw(file_path)
        raw = set_up_raw(raw, rerefer=rerefer, chan_limit=chan_limit, drop_chan=ch_drop,
                         montage_temp=montage_temp, is_montage=is_montage)

        stage = None
        if os.path.isfile(os.path.join(stage_folder, f"{file_name}.mat")):
            stage = hypnogram_segment(raw=raw, interval=scoring_dur,
                                      file_paths=os.path.join(stage_folder, f"{file_name}.mat"))
        elif os.path.isfile(os.path.join(stage_folder, f"{file_name}.txt")):
            stage = hypnogram_segment(raw=raw, interval=scoring_dur,
                                      file_paths=os.path.join(stage_folder, f"{file_name}.txt"))
        else:
            print(f"Neither {file_name}.mat nor {file_name}.txt exists in the folder.")

        event_summary, coupling_event, so_waveform = get_sosp(raw, stage,
                                                              interest_stage=interest_stage, sp_method=sp_method,
                                                              so_method=so_method, coupling=coupling,
                                                              filter_freq=filter_freq, duration=duration,
                                                              filter_type=filter_type, l_freq=l_freq,
                                                              h_freq=h_freq, dur_lower=dur_lower, dur_upper=dur_upper,
                                                              baseline=baseline, verbose=verbose)

        event_summary_all[file_name] = event_summary
        coupling_event_all[file_name] = coupling_event
        so_waveform_all[file_name] = so_waveform
    return event_summary_all, coupling_event_all, so_waveform_all



import mne
import numpy as np
import scipy.signal
import math
import pandas as pd
from mne.io import Raw
from scipy.signal import hilbert


def SO_detection(raw, stage, target_stage=('N2', 'SWS'), filter_freq=None, duration=None, baseline=True,
                 filter_type='fir',
                 method='Staresina', verbose=True):
    method_params = {
        'Staresina': {'filter_freq': (0.16, 1.25), 'duration': (0.8, 2)},
        'Ng': {'filter_freq': (0.16, 1.25), 'duration': (0.8, 2)},
        'Massimini': {'filter_freq': (0.1, 4), 'duration': (0.8, 2)},
        'Dimulescu': {'filter_freq': (0.1, 4), 'duration': (0.8, 2)},
        'Molle': {'filter_freq': (0.1, 2), 'duration': (0.9, 2)},
        'Ngo': {'filter_freq': (None, 3.5), 'duration': (0.9, 2)}
    }

    stage_mapping = {"Wake": 0, "N1": 1, "N2": 2, "SWS": 3, "REM": 4}
    target_stage = [
        stage_mapping[stg] if isinstance(stg, str) and stg in stage_mapping else stg
        for stg in target_stage
    ]
    defaults = method_params.get(method, {'filter_freq': (0.25, 3), 'duration': (0.8, 2)})
    filter_freq = filter_freq or defaults['filter_freq']
    duration = duration or defaults['duration']

    l_freq, h_freq = filter_freq
    dur_lower, dur_upper = duration

    # Drop unnecessary channels
    drop_ch = [ch for ch in raw.ch_names if ch.startswith('M') or any(x in ch for x in ['EMG', 'EOG', 'ECG'])]
    raw_copy = raw.copy().drop_channels(drop_ch)

    # Apply filtering
    filter_params = {
        'picks': 'eeg',
        'l_freq': l_freq,
        'h_freq': h_freq,
        'phase': 'zero',
        'verbose': 'ERROR'
    }

    if filter_type == 'iir':
        filter_params.update({'method': 'iir', 'iir_params': {'order': 2, 'ftype': 'butter'}})
    else:
        filter_params.update({
            'method': 'fir',
            'fir_design': 'firwin',
            'phase': 'zero-double',
            'fir_window': 'hamming',
            'filter_length': 'auto'
        })

    raw_copy.filter(**filter_params)

    eeg_data = raw_copy.get_data(picks='eeg')
    if baseline:
        eeg_centered = eeg_data - np.mean(eeg_data, axis=1, keepdims=True)
    else:
        eeg_centered = eeg_data

    annotations = raw_copy.annotations
    bad_epochs = [(ann['onset'], ann['onset'] + ann['duration']) for ann in annotations if
                  'Bad_epoch' in ann['description']]

    for bad_start, bad_end in bad_epochs:
        bad_start_idx = int(bad_start * raw_copy.info['sfreq'])
        bad_end_idx = int(bad_end * raw_copy.info['sfreq'])
        eeg_centered[:, bad_start_idx:bad_end_idx] = 0

    all_SO_candidates = []

    # Run separately for each target stage
    for t_stage in target_stage:
        stage_data = [s for s in stage if s[0] == t_stage]
        if not stage_data:
            continue
        durations_and_ptps = []

        # Process each sleep stage individually
        for idx, (stage_value, duration, _) in enumerate(stage):
            if stage_value != t_stage:
                continue

            start_time = np.sum([dur for _, dur, _ in stage[:idx]])
            end_time = start_time + duration

            for ch_idx in range(eeg_centered.shape[0]):
                global_start_idx = int(start_time * raw_copy.info['sfreq'])
                global_end_idx = int(end_time * raw_copy.info['sfreq'])
                signal = eeg_centered[ch_idx, global_start_idx:global_end_idx]
                zero_crossings = np.where(np.diff(np.signbit(signal)))[0]

                p2n_crossings = []
                for crossing in zero_crossings:
                    onset = start_time + (crossing / raw_copy.info['sfreq'])
                    if signal[crossing] > 0:
                        p2n_crossings.append(onset)

                for i in range(1, len(p2n_crossings)):
                    p2n_dur = p2n_crossings[i] - p2n_crossings[i - 1]

                    if dur_lower <= p2n_dur <= dur_upper:
                        start_time_global = p2n_crossings[i - 1]
                        end_time_global = p2n_crossings[i]

                        start_idx = int(start_time_global * raw_copy.info['sfreq'])
                        end_idx = int(end_time_global * raw_copy.info['sfreq'])

                        ptp_amplitude = np.ptp(eeg_centered[ch_idx, start_idx:end_idx])
                        trough_idx = np.argmin(eeg_centered[ch_idx, start_idx:end_idx])
                        trough_amplitude = eeg_centered[ch_idx, start_idx + trough_idx]

                        durations_and_ptps.append({
                            'ch_id': ch_idx + 1,
                            'ch_name': raw_copy.ch_names[ch_idx],
                            'p2n_start': start_time_global,
                            'p2n_end': end_time_global,
                            'duration': p2n_dur,
                            'ptp_amplitude': ptp_amplitude,
                            'trough_amplitude': trough_amplitude
                        })

        df = pd.DataFrame(durations_and_ptps)

        if method in ['Staresina', 'Ng']:

            if method == 'Staresina':
                top_25_pct_ptp_threshold = df['ptp_amplitude'].quantile(0.75)
                top_durations = df[df['ptp_amplitude'] >= top_25_pct_ptp_threshold]
            elif method == 'Ng':
                top_25_pct_ptp_threshold = df['ptp_amplitude'].quantile(0.7)
                top_25_pct_trough_threshold = df['trough_amplitude'].quantile(0.3)
                top_durations = df[(df['ptp_amplitude'] >= top_25_pct_ptp_threshold) & (
                        df['trough_amplitude'] <= top_25_pct_trough_threshold)]

            sfreq = raw_copy.info['sfreq']
            SO_candidate = []

            for idx, row in top_durations.iterrows():
                start_idx = int(row['p2n_start'] * sfreq)
                end_idx = int(row['p2n_end'] * sfreq)

                signal_segment = eeg_centered[row['ch_id'] - 1, start_idx:end_idx]
                trough_idx = np.argmin(signal_segment)
                trough_time = row['p2n_start'] + (trough_idx / sfreq)

                SO_candidate.append({
                    'ch_id': row['ch_id'],
                    'ch_name': row['ch_name'],
                    'p2n_start': row['p2n_start'],
                    'p2n_end': row['p2n_end'],
                    'duration': row['duration'],
                    'ptp_amplitude': row['ptp_amplitude'],
                    'trough_time': trough_time,
                    'trough_amplitude': row['trough_amplitude'],
                    'stage': t_stage  # Add stage info
                })

        elif method in ['Dimulescu', 'Massimini']:
            # Absolute method
            sfreq = raw_copy.info['sfreq']
            SO_candidate = []

            if method == 'Dimulescu':
                trough_amp_thres = -0.00004
                ptp_amp_thres = 0.00007
            elif method == 'Massimini':
                trough_amp_thres = -0.00008
                ptp_amp_thres = 0.00014

            for idx, row in df.iterrows():
                if row['trough_amplitude'] < trough_amp_thres and row['ptp_amplitude'] > ptp_amp_thres:
                    start_idx = int(row['p2n_start'] * sfreq)
                    end_idx = int(row['p2n_end'] * sfreq)

                    signal_segment = eeg_centered[row['ch_id'] - 1, start_idx:end_idx]
                    trough_idx = np.argmin(signal_segment)
                    trough_time = row['p2n_start'] + (trough_idx / sfreq)

                    SO_candidate.append({
                        'ch_id': row['ch_id'],
                        'ch_name': row['ch_name'],
                        'p2n_start': row['p2n_start'],
                        'p2n_end': row['p2n_end'],
                        'duration': row['duration'],
                        'ptp_amplitude': row['ptp_amplitude'],
                        'trough_time': trough_time,
                        'trough_amplitude': row['trough_amplitude'],
                        'stage': t_stage  # Add stage info
                    })

        elif method in ['Molle', 'Ngo']:
            # Ratio method
            sfreq = raw_copy.info['sfreq']
            SO_candidate = []

            mean_trough_amplitude = df['trough_amplitude'].mean()
            mean_ptp_amplitude = df['ptp_amplitude'].mean()

            if method == 'Molle':
                trough_ratio = 2 / 3
                ptp_ratio = 2 / 3
            elif method == 'Ngo':
                trough_ratio = 1.25
                ptp_ratio = 1.25

            for idx, row in df.iterrows():
                if row['trough_amplitude'] <= trough_ratio * mean_trough_amplitude and row[
                    'ptp_amplitude'] >= ptp_ratio * mean_ptp_amplitude:
                    start_idx = int(row['p2n_start'] * sfreq)
                    end_idx = int(row['p2n_end'] * sfreq)

                    signal_segment = eeg_centered[row['ch_id'] - 1, start_idx:end_idx]
                    trough_idx = np.argmin(signal_segment)
                    trough_time = row['p2n_start'] + (trough_idx / sfreq)

                    SO_candidate.append({
                        'ch_id': row['ch_id'],
                        'ch_name': row['ch_name'],
                        'p2n_start': row['p2n_start'],
                        'p2n_end': row['p2n_end'],
                        'duration': row['duration'],
                        'ptp_amplitude': row['ptp_amplitude'],
                        'trough_time': trough_time,
                        'trough_amplitude': row['trough_amplitude'],
                        'stage': t_stage  # Add stage info
                    })

        all_SO_candidates.append(pd.DataFrame(SO_candidate))

    final_SO_candidates = pd.concat(all_SO_candidates, ignore_index=True)

    SO_summary = []

    for t_stage in target_stage:
        stage_data = [s for s in stage if s[0] == t_stage]
        if not stage_data:
            continue
        stage_SO_candidates = final_SO_candidates[final_SO_candidates['stage'] == t_stage]

        mean_ptp_amp = stage_SO_candidates['ptp_amplitude'].mean()
        mean_dur = stage_SO_candidates['duration'].mean()
        total_dur_stage = sum([dur for stage_val, dur, _ in stage if stage_val == t_stage])
        num_segments = total_dur_stage // 30
        num_channels = eeg_centered.shape[0]
        SO_density = len(stage_SO_candidates) / (num_channels * num_segments) if num_segments > 0 else 0

        stage_mapping = {0: "Wake", 1: "N1", 2: "N2", 3: "SWS", 4: "REM"}
        if verbose:
            print(f"Detected SOs in stage {stage_mapping.get(t_stage, t_stage)}")
            print(f"Method: {method}")
            print(f"PTP amplitude: {mean_ptp_amp * 1e6:.4f}")
            print(f"Mean Duration: {mean_dur:.4f}")
            print(f"SO Density per 30s: {SO_density:.4f}")
            print(f"Total Count: {len(stage_SO_candidates)}")
            print("---------------------------------------------------------------")

        SO_summary.append({
            'stage': t_stage,
            'stage_dur': sum([d for (stg, d, _) in stage if stg == t_stage]),
            'mean_SO_amp': mean_ptp_amp,
            'mean_SO_dur': mean_dur,
            'SO_density': SO_density
        })

    SO_summary_df = pd.DataFrame(SO_summary)

    if not final_SO_candidates.empty:
        all_stage_dur = sum([d for (stg, d, _) in stage if stg in target_stage])
        num_segments_all = int(all_stage_dur // 30)
        num_channels = eeg_centered.shape[0]

        mean_SO_amp_all = final_SO_candidates['ptp_amplitude'].mean()
        mean_SO_dur_all = final_SO_candidates['duration'].mean()
        SO_density_all = len(final_SO_candidates) / (num_channels * num_segments_all) if num_segments_all > 0 else 0

        all_summary = {
            'stage': 'all',
            'stage_dur': all_stage_dur,
            'mean_SO_amp': mean_SO_amp_all,
            'mean_SO_dur': mean_SO_dur_all,
            'SO_density': SO_density_all
        }

        SO_summary_df = pd.concat([SO_summary_df, pd.DataFrame([all_summary])], ignore_index=True)

    channel_summary_list = []
    for t_stage in target_stage:
        stage_candidates = final_SO_candidates[final_SO_candidates['stage'] == t_stage]
        if stage_candidates.empty:
            continue
        total_dur_stage = sum([d for (stg, d, _) in stage if stg == t_stage])
        num_segments = total_dur_stage // 30
        for ch in stage_candidates['ch_name'].unique():
            ch_events = stage_candidates[stage_candidates['ch_name'] == ch]
            mean_amp = ch_events['ptp_amplitude'].mean()
            mean_dur = ch_events['duration'].mean()
            density = len(ch_events) / num_segments if num_segments > 0 else 0
            channel_summary_list.append({
                'stage': t_stage,
                'channel': ch,
                'stage_dur': total_dur_stage,
                'mean_SO_amp': mean_amp,
                'mean_SO_dur': mean_dur,
                'SO_density': density
            })
    channel_summary_df = pd.DataFrame(channel_summary_list)
    SO_summary_df['channel'] = 'all'
    SO_summary_df = pd.concat([SO_summary_df, channel_summary_df], ignore_index=True)

    # Add channel-level data for the aggregated 'all' stage
    if not final_SO_candidates.empty:
        channel_summary_all = []
        total_all_dur = sum([d for (stg, d, _) in stage if stg in target_stage])
        num_segments_all = total_all_dur // 30
        for ch in final_SO_candidates['ch_name'].unique():
            ch_events = final_SO_candidates[final_SO_candidates['ch_name'] == ch]
            mean_amp = ch_events['ptp_amplitude'].mean()
            mean_dur = ch_events['duration'].mean()
            density = len(ch_events) / num_segments_all if num_segments_all > 0 else 0
            channel_summary_all.append({
                'stage': 'all',
                'channel': ch,
                'stage_dur': total_all_dur,
                'mean_SO_amp': mean_amp,
                'mean_SO_dur': mean_dur,
                'SO_density': density
            })
        channel_summary_all_df = pd.DataFrame(channel_summary_all)
        SO_summary_df = pd.concat([SO_summary_df, channel_summary_all_df], ignore_index=True)

    column_order = ['stage', 'channel', 'stage_dur', 'mean_SO_amp', 'mean_SO_dur', 'SO_density']
    SO_summary_df = SO_summary_df[column_order]

    return raw_copy, final_SO_candidates, SO_summary_df


def SP_detection(raw: Raw, stage, target_stage=('N2', 'SWS'), method: str = "Hahn2020",
                 l_freq: float = 10, h_freq: float = 16, dur_lower: float = 0.5, dur_upper: float = math.inf,
                 baseline: bool = True, verbose: bool = True):
    """
    Detect sleep spindles in EEG data using five published methods:

    1) hahn2020 (default)
       - 12-16 Hz bandpass filter, Hilbert transform to obtain amplitude envelope,
         thresholded at a given percentile (default 75th),
         with a duration criterion of 0.5 to 3 seconds.

    2) martin2013
       - 11-15 Hz bandpass filter (linear phase FIR, forward and reverse for 0-phase distortion).
       - Compute RMS of the filtered signal with a 0.25 s time window.
       - Threshold at its 95th percentile (published criteria: Schabus et al., 2007).
       - A spindle is identified if at least two consecutive RMS time points
         exceed the threshold, and total duration >= 0.5 s.

    3) wamsley2012
       - Discrete sleep spindle events detected using a wavelet-based algorithm
         with an 8-parameter complex Morlet wavelet, ~10-16 Hz range.
       - Spindles are identified at each channel by a thresholding algorithm
         applied to the wavelet scale corresponding to 10-16 Hz.
       - The rectified moving average of this wavelet signal is calculated
         with a 100 ms sliding window.
       - Threshold = 4.5 times the mean signal amplitude across artifact-free epochs.
       - Duration >= 0.4 s to count as a spindle.

    4) Wendt2012
       - The raw data are bandpass filtered between 11-16 Hz.
       - The signal is then rectified.
       - Two separate lowpass filters are applied to create two envelopes
         (one ~2.25 Hz passband with offset 3 μV, the other ~1 Hz passband with offset 8 μV).
       - A time-varying threshold is formed by each envelope + offset.
       - Any rectified signal crossing either threshold boundary within 0.5-3 s
         is considered a spindle candidate. A final decision fusion is applied

    5) Ferrarelli2007
       - 12-15 Hz bandpass filter (–3 dB at 12 and 15 Hz).
       - The amplitude of the rectified filtered signal is used as the time series.
       - If the signal's amplitude exceeds an upper threshold (8 × channel average amplitude),
         it is considered a spindle candidate. Peak amplitude is the local maximum above
         that threshold. The start and end are the preceding/following points
         where amplitude drops below a lower threshold (2 × channel average amplitude).
       - Duration >= 0.5 s is typically required (AASM-based).

    Parameters
    ----------
    verbose
    raw : mne.io.Raw
        MNE Raw object containing EEG data.
    stage : list of tuples
        A list describing sleep stage segments, e.g. [(stage_value, dur_sec, label), ...].
        The function accumulates these durations in order to determine
        the time boundaries for each stage segment.
    method : str
        The spindle detection method to be used. Must be one of:
        ["hahn2020", "martin2013", "wamsley2012", "Wendt2012", "Ferrarelli2007"].
    l_freq : float
        Default low cut-off frequency for the bandpass filter
    h_freq : float
        Default high cut-off frequency for the bandpass filter
    dur_lower : float
        Minimum duration (seconds) for a valid spindle event.
    dur_upper : float
        Maximum duration (seconds) for a valid spindle event.
    baseline : bool
        Whether to subtract the mean of the entire channel signal
        from each channel prior to detection.
    target_stage : list
        List of stage values (e.g., [2, 3]) on which detection will be performed.

    Returns
    -------
    raw_copy : mne.io.Raw
        A copy of the Raw object with only EEG channels of interest
        (e.g., excluding EMG/EOG/ECG), filtered according to the chosen method.
    final_SP_candidates : pd.DataFrame
        A DataFrame containing all detected spindle events across channels
        and target stages, with columns including 'ch_id', 'ch_name',
        'start_time', 'end_time', 'peak_time', 'duration', 'amplitude', etc.
    """

    raw_copy = raw.copy()
    sfreq = raw_copy.info['sfreq']
    stage_mapping = {"Wake": 0, "N1": 1, "N2": 2, "SWS": 3, "REM": 4}
    target_stage = [
        stage_mapping[stg] if isinstance(stg, str) and stg in stage_mapping else stg
        for stg in target_stage
    ]
    if method == "Hahn2020":
        # 12-16 Hz
        bp_l_freq = 12
        bp_h_freq = 16

    elif method == "Martin2013":
        # 11-15 Hz
        bp_l_freq = 11
        bp_h_freq = 15

    elif method == "Wamsley2012":
        # wavelet-based approach (10-16 Hz scale).
        # We will still do an initial bandpass (approx 10-16) to remove out-of-band noise
        # prior to wavelet transform if desired, but the main detection is wavelet-based.
        bp_l_freq = 10
        bp_h_freq = 16

    elif method == "Wendt2012":
        # 11-16 Hz
        bp_l_freq = 11
        bp_h_freq = 16

    elif method == "Ferrarelli2007":
        # 12-15 Hz
        bp_l_freq = 12
        bp_h_freq = 15

    else:
        bp_l_freq = l_freq
        bp_h_freq = h_freq

    raw_copy.filter(
        picks='eeg',
        l_freq=bp_l_freq,
        h_freq=bp_h_freq,
        method='fir',
        fir_design='firwin',
        phase='zero-double',
        fir_window='hamming',
        filter_length='auto',
        verbose='ERROR'
    )

    eeg_data = raw_copy.get_data(picks='eeg')
    if baseline:
        eeg_centered = eeg_data - np.mean(eeg_data, axis=1, keepdims=True)
    else:
        eeg_centered = eeg_data

    annotations = raw_copy.annotations
    bad_epochs = [
        (ann['onset'], ann['onset'] + ann['duration'])
        for ann in annotations
        if 'Bad_epoch' in ann['description']
    ]
    for bad_start, bad_end in bad_epochs:
        bad_start_idx = int(bad_start * sfreq)
        bad_end_idx = int(bad_end * sfreq)
        eeg_centered[:, bad_start_idx:bad_end_idx] = 0

    # 8. Loop through target stages and detect spindles segment by segment
    all_stage_dfs = []
    n_channels = eeg_centered.shape[0]

    for s_val in target_stage:
        # find all segments for this stage
        stage_segments = [(val, dur, label) for (val, dur, label) in stage if val == s_val]
        if not stage_segments:
            continue

        local_candidates = []

        # Accumulate time so we know start index for each segment
        accumulated_time = 0.0
        for idx, (val, duration_s, _) in enumerate(stage):
            if val != s_val:
                accumulated_time += duration_s
                continue

            segment_start_time = accumulated_time
            segment_end_time = segment_start_time + duration_s
            accumulated_time += duration_s

            seg_start_idx = int(segment_start_time * sfreq)
            seg_end_idx = int(segment_end_time * sfreq)

            for ch_idx in range(n_channels):
                signal_seg = eeg_centered[ch_idx, seg_start_idx:seg_end_idx]

                # Detect spindles on this single-epoch segment
                sp_list = detect_spindles_for_epoch(method, signal_seg, dur_lower, dur_upper, sfreq)
                # Convert local indices to global times
                for sp in sp_list:
                    local_candidates.append({
                        'ch_id': ch_idx + 1,
                        'ch_name': raw_copy.ch_names[ch_idx],
                        'start_time': (sp['start_idx'] + seg_start_idx) / sfreq,
                        'end_time': (sp['end_idx'] + seg_start_idx) / sfreq,
                        'peak_time': (sp['peak_idx'] + seg_start_idx) / sfreq,
                        'duration': sp['duration'],
                        'amplitude': sp['amplitude'],
                        'stage': s_val
                    })

        # Convert to DataFrame
        stage_df = pd.DataFrame(local_candidates)
        if not stage_df.empty:
            # Optional Z-normalization across spindles in this stage
            stage_df['z_amplitude'] = (
                    (stage_df['amplitude'] - stage_df['amplitude'].mean()) /
                    stage_df['amplitude'].std()
            )
            stage_df['z_duration'] = (
                    (stage_df['duration'] - stage_df['duration'].mean()) /
                    stage_df['duration'].std()
            )
        all_stage_dfs.append(stage_df)

    # 9. Concatenate results from all target stages
    if all_stage_dfs:
        final_SP_candidates = pd.concat(all_stage_dfs, ignore_index=True)
    else:
        final_SP_candidates = pd.DataFrame(
            columns=['ch_id', 'ch_name', 'start_time', 'end_time',
                     'peak_time', 'duration', 'amplitude', 'stage',
                     'z_amplitude', 'z_duration']
        )

    SP_summary = []

    for s_val in target_stage:
        stage_df = final_SP_candidates[final_SP_candidates['stage'] == s_val]
        if stage_df.empty:
            continue

        if method == "Wamsley2012":
            mean_amp = np.sqrt(stage_df['amplitude'].mean())  # Convert power to amplitude
        else:
            mean_amp = stage_df['amplitude'].mean()
        mean_dur = stage_df['duration'].mean()
        total_dur_stage = sum([d for (stg, d, _) in stage if stg == s_val])
        num_segments = int(total_dur_stage // 30)
        sp_density = len(stage_df) / (n_channels * num_segments) if num_segments > 0 else 0

        stage_mapping = {0: "Wake", 1: "N1", 2: "N2", 3: "SWS", 4: "REM"}
        if verbose:
            print(f"Detected SPs in stage {stage_mapping.get(s_val, s_val)}")
            print(f"Method: {method}")
            print(f"Average Amplitude: {mean_amp * 1e6:.4f} uv")
            print(f"Average Duration: {mean_dur:.4f}s")
            print(f"Spindle Density (per 30s): {sp_density:.4f}")
            print(f"Total Spindle Count: {len(stage_df)}")
            print("---------------------------------------------------------------")

        SP_summary.append({
            'stage': s_val,
            'mean_SP_amp': mean_amp,
            'mean_SP_dur': mean_dur,
            'SP_density': sp_density
        })

    SP_summary_df = pd.DataFrame(SP_summary)

    if not final_SP_candidates.empty:
        all_stage_dur = sum([d for (stg, d, _) in stage if stg in target_stage])
        num_segments_all = int(all_stage_dur // 30)
        num_channels = eeg_centered.shape[0]

        mean_SP_amp_all = final_SP_candidates['amplitude'].mean()
        mean_SP_dur_all = final_SP_candidates['duration'].mean()
        SP_density_all = len(final_SP_candidates) / (num_channels * num_segments_all) if num_segments_all > 0 else 0

        all_summary = {
            'stage': 'all',
            'mean_SP_amp': mean_SP_amp_all,
            'mean_SP_dur': mean_SP_dur_all,
            'SP_density': SP_density_all
        }

        SP_summary_df = pd.concat([SP_summary_df, pd.DataFrame([all_summary])], ignore_index=True)

    channel_summary_list = []
    for s_val in target_stage:
        stage_candidates = final_SP_candidates[final_SP_candidates['stage'] == s_val]
        if stage_candidates.empty:
            continue
        total_dur_stage = sum([d for (stg, d, _) in stage if stg == s_val])
        num_segments = int(total_dur_stage // 30)
        for ch in stage_candidates['ch_name'].unique():
            ch_events = stage_candidates[stage_candidates['ch_name'] == ch]
            mean_amp = ch_events['amplitude'].mean()
            mean_dur = ch_events['duration'].mean()
            density = len(ch_events) / num_segments if num_segments > 0 else 0
            channel_summary_list.append({
                'stage': s_val,
                'channel': ch,
                'mean_SP_amp': mean_amp,
                'mean_SP_dur': mean_dur,
                'SP_density': density
            })
    channel_summary_df = pd.DataFrame(channel_summary_list)
    SP_summary_df['channel'] = 'all'
    SP_summary_df = pd.concat([SP_summary_df, channel_summary_df], ignore_index=True)

    # Add channel-level data for the aggregated 'all' stage
    if not final_SP_candidates.empty:
        channel_summary_all = []
        total_all_dur = sum([d for (stg, d, _) in stage if stg in target_stage])
        num_segments_all = int(total_all_dur // 30)
        for ch in final_SP_candidates['ch_name'].unique():
            ch_events = final_SP_candidates[final_SP_candidates['ch_name'] == ch]
            mean_amp = ch_events['amplitude'].mean()
            mean_dur = ch_events['duration'].mean()
            density = len(ch_events) / (num_segments_all) if num_segments_all > 0 else 0
            channel_summary_all.append({
                'stage': 'all',
                'channel': ch,
                'mean_SP_amp': mean_amp,
                'mean_SP_dur': mean_dur,
                'SP_density': density
            })
        channel_summary_all_df = pd.DataFrame(channel_summary_all)
        SP_summary_df = pd.concat([SP_summary_df, channel_summary_all_df], ignore_index=True)

    column_order = ['stage', 'channel', 'mean_SP_amp', 'mean_SP_dur', 'SP_density']
    SP_summary_df = SP_summary_df[column_order]

    return raw_copy, final_SP_candidates, SP_summary_df


def swa(raw: Raw, file_name = 'id', l_freq=0.5, h_freq=4):
    """
    Compute slow wave activity (SWA) from an mne.Raw object.

    This function filters the raw EEG data in the slow wave band (0.5-4 Hz),
    computes the Hilbert transform to extract the amplitude envelope,
    and then averages this envelope over time for each channel.

    Parameters:
    -----------
    raw : mne.io.Raw
        The raw EEG/MEG data.

    Returns:
    --------
    pd.DataFrame
        A DataFrame containing the slow wave activity for each channel.
    """
    # Sampling frequency
    fs = raw.info['sfreq']
    # Channel names
    ch_names = raw.info['ch_names']
    # Get the data as a numpy array (shape: n_channels x n_times)
    data = raw.get_data()

    # Bandpass filter the data to the slow wave (delta) band: 0.5 - 4 Hz
    filtered_data = mne.filter.filter_data(data, sfreq=fs, l_freq=l_freq, h_freq=h_freq, verbose=False)

    # Compute the analytic signal via Hilbert transform along the time axis
    analytic_signal = scipy.signal.hilbert(filtered_data, axis=1)
    # Compute the amplitude envelope (i.e. the instantaneous amplitude)
    amplitude_envelope = np.abs(analytic_signal)

    # Average the amplitude envelope over time for each channel to get a SWA measure
    swa_values = np.mean(amplitude_envelope, axis=1)

    # # Create and return a DataFrame with the results
    # df = pd.DataFrame({
    #     'Channel': ch_names,
    #     'Slow_Wave_Activity': swa_values
    # })
    # Create a dictionary mapping each channel name to its slow wave activity value.
    data_dict = {ch: swa for ch, swa in zip(ch_names, swa_values)}

    # Create a DataFrame with one row using the filename as the index.
    df = pd.DataFrame([data_dict])

    # Optionally, insert the filename as the first column.
    df.insert(0, 'Participant_id', file_name)

    return df


def detect_spindles_for_epoch(method, signal_1d, dur_lower, dur_upper, sfreq):
    # 7. Define helper function to detect spindles from a single epoch signal
    #    according to the chosen method
    """
    Detect spindle events in one single-epoch, single-channel signal
    using the specified 'method'.
    Returns a list of dictionaries (one per spindle event).
    """
    spindle_candidates = []

    if method == "Hahn2020":
        spindle_candidates = hahn(sfreq, signal_1d)

    elif method == "Martin2013":
        spindle_candidates = martin(sfreq, signal_1d)

    elif method == "Wamsley2012":
        spindle_candidates = wamsley(sfreq, signal_1d)

    elif method == "Wendt2012":
        spindle_candidates = wendt(sfreq, signal_1d)

    elif method == "Ferrarelli2007":
        spindle_candidates = ferrarelli(dur_lower, sfreq, signal_1d)

    else:
        # Fallback to Hahn2020 approach if method is unknown
        analytic_signal = hilbert(signal_1d)
        amplitude_envelope = np.abs(analytic_signal)
        window_samples = int(0.2 * sfreq)
        smoothed_envelope = np.convolve(
            amplitude_envelope,
            np.ones(window_samples) / window_samples,
            mode='same'
        )
        amp_threshold = np.percentile(smoothed_envelope, 75)

        above = smoothed_envelope > amp_threshold
        start_idx = None
        for i, val in enumerate(above):
            if val and start_idx is None:
                start_idx = i
            elif (not val) and (start_idx is not None):
                end_idx = i
                duration_sec = (end_idx - start_idx) / sfreq
                if dur_lower <= duration_sec <= dur_upper:
                    peak_local = np.argmax(smoothed_envelope[start_idx:end_idx]) + start_idx
                    peak_amp = smoothed_envelope[peak_local]
                    spindle_candidates.append({
                        'start_idx': start_idx,
                        'end_idx': end_idx,
                        'peak_idx': peak_local,
                        'duration': duration_sec,
                        'amplitude': peak_amp
                    })
                start_idx = None

    return spindle_candidates


def ferrarelli(dur_lower, sfreq, signal_1d):
    # (Ferrarelli 2007)
    # 12–15 Hz bandpass (already done), rectify signal -> amplitude time series.
    # Upper threshold = 8 × mean amplitude => potential spindle region starts there.
    # Then continues until amplitude < lower threshold = 2 × mean amplitude.
    spindle_candidates = []

    rectified = np.abs(signal_1d)
    mean_amp = np.mean(rectified)
    upper_thr = 8.0 * mean_amp
    lower_thr = 2.0 * mean_amp
    in_spindle = False
    start_idx = None
    for i, val in enumerate(rectified):
        if (not in_spindle) and (val > upper_thr):
            in_spindle = True
            start_idx = i
        elif in_spindle and (val < lower_thr):
            end_idx = i
            duration_sec = (end_idx - start_idx) / sfreq
            if duration_sec >= dur_lower:
                peak_local = np.argmax(rectified[start_idx:end_idx]) + start_idx
                peak_amp = rectified[peak_local]
                spindle_candidates.append({
                    'start_idx': start_idx,
                    'end_idx': end_idx,
                    'peak_idx': peak_local,
                    'duration': duration_sec,
                    'amplitude': peak_amp
                })
            in_spindle = False

    return spindle_candidates


def wendt(sfreq, signal_1d):
    spindle_candidates = []
    rectified = np.abs(signal_1d)
    # Envelope #1 (approx passband 2.25 Hz) -> e.g. ~0.45s kernel
    win_size1 = int(0.45 * sfreq)
    envelope1 = np.convolve(rectified, np.ones(win_size1) / win_size1, mode='same')
    offset1 = 3.0  # in μV
    # Envelope #2 (approx passband 1 Hz) -> ~1s kernel
    win_size2 = int(1.0 * sfreq)
    envelope2 = np.convolve(rectified, np.ones(win_size2) / win_size2, mode='same')
    offset2 = 8.0  # in μV
    thresh_array = np.minimum(envelope1 + offset1, envelope2 + offset2)
    above = rectified > thresh_array
    start_idx = None
    for i, val in enumerate(above):
        if val and start_idx is None:
            start_idx = i
        elif (not val) and (start_idx is not None):
            end_idx = i
            duration_sec = (end_idx - start_idx) / sfreq
            if 0.5 <= duration_sec <= 3:
                peak_local = np.argmax(rectified[start_idx:end_idx]) + start_idx
                peak_amp = rectified[peak_local]
                spindle_candidates.append({
                    'start_idx': start_idx,
                    'end_idx': end_idx,
                    'peak_idx': peak_local,
                    'duration': duration_sec,
                    'amplitude': peak_amp
                })
            start_idx = None

    return spindle_candidates


def wamsley(sfreq, signal_1d):
    # (Wamsley 2012)
    # Discrete sleep spindle events automatically detected via a wavelet-based algorithm
    # covering ~10-16 Hz. For thresholding, the rectified moving average (100 ms window)
    # of the wavelet scale is used. A spindle is identified if the signal
    # exceeds 4.5 * mean amplitude for >= 400 ms.
    #
    # Implementation note:
    # We'll use MNE's Morlet wavelet transform for the relevant frequencies (10-16 Hz),
    # combine power or amplitude across that band, rectify, and smooth with 100 ms.
    #
    # 1. Create a short "dummy" RawArray for wavelet analysis. Or we can do TFR in a direct manner.
    # 2. We'll compute time-frequency representation, sum (or average) power in 10-16 Hz range.
    # 3. Rectify => take the absolute value (or it's already power). Then a 100 ms moving average.
    # 4. threshold = 4.5 * mean
    # 5. events must last >= 0.4 s.
    spindle_candidates = []
    # We'll shape the data into (n_epochs=1, n_channels=1, n_times) for mne tfr_morlet:
    signal_2d = signal_1d[np.newaxis, np.newaxis, :]
    # freq range
    freqs = np.arange(10, 17, 1)  # 10 to 16 Hz inclusive
    # We'll use ~8 cycles for Morlet, based on "8-parameter complex Morlet wavelet"
    n_cycles = 8
    # Create an info just for this single channel
    info_tmp = mne.create_info(ch_names=['temp'], sfreq=sfreq, ch_types=['eeg'])
    raw_tmp = mne.io.RawArray(signal_2d[0], info_tmp, verbose=False)
    # Now compute TFR with Morlet
    # We want a single-epoch structure, so create a dummy epoch
    power_data = mne.time_frequency.tfr_array_morlet(
        signal_1d[np.newaxis, np.newaxis, :],  # (1, 1, n_times)
        sfreq=sfreq,
        freqs=freqs,
        n_cycles=n_cycles,
        output='power',
        use_fft=True,
        decim=1,
        verbose=False
    )[0, 0]
    avg_power = np.mean(power_data, axis=0)
    window_samples = int(0.1 * sfreq)
    smoothed = np.convolve(avg_power, np.ones(window_samples) / window_samples, mode='same')
    thr = 4.5 * np.mean(smoothed)
    above = smoothed > thr
    start_idx = None
    for i, val in enumerate(above):
        if val and start_idx is None:
            start_idx = i
        elif (not val) and (start_idx is not None):
            end_idx = i
            duration_sec = (end_idx - start_idx) / sfreq
            if 0.4 <= duration_sec:
                peak_local = np.argmax(smoothed[start_idx:end_idx]) + start_idx
                peak_amp = smoothed[peak_local]
                spindle_candidates.append({
                    'start_idx': start_idx,
                    'end_idx': end_idx,
                    'peak_idx': peak_local,
                    'duration': duration_sec,
                    'amplitude': peak_amp
                })
            start_idx = None

    return spindle_candidates


def martin(sfreq, signal_1d):
    # (Martin et al. 2013, referencing Schabus et al. 2007)
    # 11-15 Hz bandpass (already done).
    # Then, compute RMS using a 0.25-s window, threshold at 95th percentile.
    spindle_candidates = []
    window_samples = int(0.25 * sfreq)
    squared_signal = signal_1d ** 2
    # moving average of squared => RMS
    moving_avg = np.convolve(
        squared_signal,
        np.ones(window_samples) / window_samples,
        mode='same'
    )
    rms_envelope = np.sqrt(moving_avg)
    amp_threshold = np.percentile(rms_envelope, 95)
    above = rms_envelope > amp_threshold
    start_idx = None
    for i, val in enumerate(above):
        if val and start_idx is None:
            start_idx = i
        elif (not val) and (start_idx is not None):
            end_idx = i
            duration_sec = (end_idx - start_idx) / sfreq
            # Martin2013 requires at least 0.5 s
            if duration_sec >= 0.5:
                peak_local = np.argmax(rms_envelope[start_idx:end_idx]) + start_idx
                peak_amp = rms_envelope[peak_local]
                spindle_candidates.append({
                    'start_idx': start_idx,
                    'end_idx': end_idx,
                    'peak_idx': peak_local,
                    'duration': duration_sec,
                    'amplitude': peak_amp
                })
            start_idx = None

    return spindle_candidates


def hahn(sfreq, signal_1d):
    # (Hahn et al. 2020)
    # Hilbert transform -> extract amplitude envelope thresholded at a chosen percentile
    # default percentile=75, duration 0.5-3 s
    spindle_candidates = []
    analytic_signal = hilbert(signal_1d)
    amplitude_envelope = np.abs(analytic_signal)
    # Smooth with a 200 ms moving average
    window_samples = int(0.2 * sfreq)
    smoothed_envelope = np.convolve(
        amplitude_envelope,
        np.ones(window_samples) / window_samples,
        mode='same'
    )
    amp_threshold = np.percentile(smoothed_envelope, 75)
    above = smoothed_envelope > amp_threshold
    start_idx = None
    consecutive_count = 0  # Count consecutive points above threshold
    for i, val in enumerate(above):
        if val:
            if start_idx is None:
                start_idx = i
            consecutive_count += 1
        elif start_idx is not None:
            if consecutive_count >= 2:  # Ensure at least 2 consecutive points
                end_idx = i
                duration_sec = (end_idx - start_idx) / sfreq
                if 0.5 <= duration_sec <= 3:
                    peak_local = np.argmax(smoothed_envelope[start_idx:end_idx]) + start_idx
                    peak_amp = smoothed_envelope[peak_local]
                    spindle_candidates.append({
                        'start_idx': start_idx,
                        'end_idx': end_idx,
                        'peak_idx': peak_local,
                        'duration': duration_sec,
                        'amplitude': peak_amp
                    })
            # Reset for the next segment
            start_idx = None
            consecutive_count = 0

    return spindle_candidates
